/* cmod - set up the user environment with module files.
   Copyright (C) 1997 by the authors of cmod (see AUTHORS).

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "env.h"
#include "cmod_dirent.h"
#include "cmod_main.h"
#include "cmod_trace.h"
#include "cmod_parse.h"
#include "cmod_alloc.h"
#include "unused.h"
#include "archaic.h"

/* Struct for "module avail" */
struct avail_info {
  char *name;			/* Includes the directory, if any. */
  int is_default;		/* True if this module is the default
				   version of a module that is a
				   directory. */
};


/* Forward declarations. */
#ifdef PROTOTYPES
static void parse_module(FILE *fp,
			 const char *filename,
			 int for_removal,
			 int *lineno);
#else
static void parse_module();
#endif

/* Random globals */

const char *argv0 = NULL;
int cmod_errors = 0;
struct envarray *global_env = NULL;
static int warning_found = 0;


#define DEFAULT_MODULE "default"


static void
warning(const char *filename,
	const int *lineno,
	const char *msg)
{
  warning_found = 1;
  if (filename != NULL && lineno != NULL)
    fprintf(stderr, "%s:%d: warning: %s\n", filename, *lineno, msg);
  else
    fprintf(stderr, "module: warning: %s\n", msg);
}


static int
append_path(const char *UNUSED(module),
	    const int *UNUSED(line),
	    int argc,
	    const char *const *argv)
{
  const char *name;
  const char *frags;
  struct env *e;

  assert(argc == 2);
  name = argv[0];
  frags = argv[1];

  e = envarray_locate_create(global_env, name, 1);
  env_append_frags(e, ':', frags, 1);
  return 0;
}


static int
prepend_path(const char *UNUSED(module),
	     const int *UNUSED(line),
	     int argc,
	     const char *const *argv)
{
  const char *name;
  const char *frags;
  struct env *e;

  assert(argc == 2);
  name = argv[0];
  frags = argv[1];

  e = envarray_locate_create(global_env, name, 1);
  env_prepend_frags(e, ':', frags, 1);
  return 0;
}


static int
remove_path(const char *UNUSED(module),
	    const int *UNUSED(line),
	    int argc,
	    const char *const *argv)
{
  const char *name;
  const char *frags;
  struct env *e;
  
  assert(argc == 2);

  name = argv[0];
  frags = argv[1];

  e = envarray_locate_create(global_env, name, 1);
  env_remove_frags(e, ':', frags);
  return 0;
}


static int
set_env(const char *UNUSED(module),
	const int *UNUSED(line),
	int argc,
	const char *const *argv)
{
  const char *name;
  char *val;
  struct env *e;

  assert(argc == 2);

  name = argv[0];
  val = xstrdup(argv[1]);

  e = envarray_locate_create(global_env, name, 0);
  env_setenv(e, val);
  return 0;
}


static int
unset_env(const char *UNUSED(module),
	  const int *UNUSED(line),
	  int argc,
	  const char *const *argv)
{
  const char *name;
  struct env *e;

  assert(argc == 2 || argc == 1);

  name = argv[0];

  e = envarray_locate_create(global_env, name, 0);
  env_unsetenv(e);
  return 0;
}


static void
do_use_path(const char *path)
{
  struct env *e = envarray_locate_create(global_env, "MODULEPATH", 1);
  env_prepend_frags(e, ':', path, 1);
}

static int
use_path(const char *UNUSED(containing_filename),
	 const int *UNUSED(containing_lineno),
	 int argc,
	 const char *const *argv)
{
  assert(argc == 1);
  do_use_path(argv[0]);
  return 0;
}


static void
do_unuse_path(const char *path)
{
  struct env *e = envarray_locate_create(global_env, "MODULEPATH", 1);
  env_remove_frags(e, ':', path);
}

static int
unuse_path(const char *UNUSED(containing_filename),
	 const int *UNUSED(containing_lineno),
	 int argc,
	 const char *const *argv)
{
  assert(argc == 1);
  do_unuse_path(argv[0]);
  return 0;
}


/* Open MODULE for reading and return a FILE*.  Search for MODULE in
   the environment variable MODULEPATH (which is stored in the global
   environment if it wasn't already present there).  The file name is
   returned in FOUND_FILENAME.  Ownership of the filename passes to
   the caller.

   If MODULE turns out to be a directory, this will open
   MODULE/default if such a file exists.  Otherwise this function will
   fail.

   Prints an error message and exits if the module cannot be found.
   FIXME: this semantics should be changed. */

static FILE *
open_module(const char *module,
	    char **found_filename)
{
  int component = 0;
  const char *dir;
  FILE *fp;
  struct env *module_path;
  char *fn;
  struct stat statbuf;
  size_t len;

  module_path = envarray_locate_create(global_env, "MODULEPATH", 1);

  while ((dir = env_iterate(module_path, ':', &component)) != NULL)
    {
      /* We allocate for the worst case at once:
	 strlen(dir)    -- the directory
         1              -- one slash
	 strlen(module) -- the module name
	 1              -- one slash
	 sizeof(DEFAULT_MODULE) -- the version name and trailing NUL. */

      len = strlen(dir) + 2 + strlen(module) + sizeof(DEFAULT_MODULE);
      fn = xmalloc(len);
      sprintf(fn, "%s/%s", dir, module);
      if (stat(fn, &statbuf) == 0)
	{
	  if (S_ISDIR(statbuf.st_mode))
	    {
	      sprintf(fn + (len - sizeof(DEFAULT_MODULE) - 1), "/%s",
		      DEFAULT_MODULE);
	      if (stat(fn, &statbuf) == 0)
		{
#ifndef STAT_MACROS_BROKEN
		  if (S_ISDIR(statbuf.st_mode))
		    {
		      fprintf(stderr,
			      "%s: opening module %s: %s is a directory\n",
			      argv0, module, fn);
		      exit(CMOD_EXIT_ERROR);
		    }
#endif
		  fp = fopen(fn, "r");
		  if (fp != NULL)
		    {
		      *found_filename = fn;
		      return fp;
		    }
		  else
		    {
		      fprintf(stderr, "%s: opening ", argv0);
		      perror(fn);
		      exit(CMOD_EXIT_ERROR);
		    }
		}
	      else if (errno != ENOENT)
		{
		  fprintf(stderr, "%s: while statting ", argv0);
		  perror(fn);
		  exit(CMOD_EXIT_ERROR);
		}
	      else
		{
		  fn[len - sizeof(DEFAULT_MODULE) - 1] = '\0';
#ifdef STAT_MACROS_BROKEN
		  /* dir/module might not be a directory after all.
		     Fall through to the code below that tries to open
		     fn as a plain file.  This means that the error
		     message "dir lacks dir/default" will never be
		     displayed on hosts that with STAT_MACROS_BROKEN,
		     which is unfortunate, but at least a proper setup
		     will work on them this way.  (Actually, a similar
		     message might be displayed anyhow, since we
		     attempt to trap this error in get_word.) */
#else
		  fprintf(stderr, "%s: %s lacks %s/default\n", argv0, fn, fn);
		  exit(CMOD_EXIT_ERROR);
#endif
		}
	    }

	  /* If we get here, assume that fn contains a file name.
	     If STAT_MACROS_BROKEN is true, it may actually be a
	     directory.  We try to catch that error when we later read
	     from the file. */

	  errno = 0;
	  fp = fopen(fn, "r");
	  if (fp == NULL)
	    {
#ifdef EISDIR
	      if (errno == EISDIR)
		{
		  fprintf(stderr, "%s: %s lacks %s/default\n", argv0, fn, fn);
		  exit(CMOD_EXIT_ERROR);
		}
#endif
	      /* Strange.  We could stat this file, but we cannot open
		 it.  Maybe we don't have permission to read it. */
	      fprintf(stderr, "%s: opening ", argv0);
	      perror(fn);
	      exit(CMOD_EXIT_ERROR);
	    }
	  
	  *found_filename = fn;
	  return fp;
	}
      else if (errno != ENOENT)
	{
	  fprintf(stderr, "%s: while statting ", argv0);
	  perror(fn);
	  exit(CMOD_EXIT_ERROR);
	}
      xfree(fn);
    }

  fprintf(stderr, "%s: failed to find %s in $MODULEPATH.\n", argv0, module);
  exit(CMOD_EXIT_ERROR);
}

  
static void
do_rm_module(const char *containing_filename,
	     const int *containing_lineno,
	     const char *module_name)
{
  char *fn;
  int lineno = 0;
  FILE *mfp;

  if (env_frag_present(envarray_locate_create(global_env, "LOADEDMODULES", 1),
		       ':', module_name))
    {
      mfp = open_module(module_name, &fn);
      assert(mfp != NULL && fn != NULL);

      parse_module(mfp, fn, 1, &lineno);
      env_remove_one_frag(envarray_locate_create(global_env, "LOADEDMODULES",
						 1),
			  ':', module_name);
      fclose(mfp);
      xfree(fn);
    }
  else
    warning(containing_filename, containing_lineno,
	    "ignored attempt to unload unloaded module.");
}

static int
rm_module(const char *containing_filename,
	  const int *containing_lineno,
	  int argc,
	  const char *const *argv)
{
  int opti;

  for (opti = 0; opti < argc; ++opti)
    do_rm_module(containing_filename, containing_lineno, argv[opti]);
  return 0;
}


static void
do_add_module(const char *containing_filename,
	      const int *containing_lineno,
	      const char *module_name)
{
  int lineno = 0;
  FILE *mfp;
  char *fn;
  struct env *loaded_modules;
  
  if (env_frag_present(envarray_locate_create(global_env, "LOADEDMODULES", 1),
		       ':', module_name))
    do_rm_module(containing_filename, containing_lineno, module_name);

  mfp = open_module(module_name, &fn);
  assert(mfp != NULL && fn != NULL);

  parse_module(mfp, fn, 0, &lineno);

  loaded_modules = envarray_locate_create(global_env, "LOADEDMODULES", 1);
  env_append_one_frag(loaded_modules, ':', module_name);

  fclose(mfp);
  xfree(fn);
}


static int
add_module(const char *containing_filename,
	   const int *containing_lineno,
	   int argc,
	   const char *const *argv)
{
  int opti;

  for (opti = 0; opti < argc; ++opti)
    do_add_module(containing_filename, containing_lineno, argv[opti]);
  return 0;
}



static int
list_modules(const char *UNUSED(from_module),
	     const int *UNUSED(from_line),
	     int argc,
	     const char *const *UNUSED(argv))
{
  struct env *loaded_modules;
  int i;
  const char *module;

  assert(argc == 0);

  loaded_modules = envarray_locate_create(global_env, "LOADEDMODULES", 1);

  fprintf(stderr,"Currently loaded modules:\n");
  for (i = 0; (module = env_iterate(loaded_modules, ':', &i)) != NULL; )
    fprintf(stderr,"%3d) %s\n", i, module);

  return 0;
}

static int
no_op(const char *UNUSED(module),
      const int *UNUSED(line),
      int UNUSED(argc),
      const char *const *UNUSED(argv))
{
  return 0;
}


static int
is_module_name(const char *name)
{
  const char *end;

  /* Avoid recursion. */
  if (!strcmp(name, "."))
    return 0;
  if (!strcmp(name, ".."))
    return 0;

  end = strchr(name, '\0');

  /* Anything that ends in a tilde is probably an Emacs autosave file.
     Avoid clutter. */
  if (end - name > 0 && end[-1] == '~')
    return 0;

  return 1;
}

static void
avail_modules_recursively(const char *dir,
			  const char *prefix,
			  const char *default_module,
			  struct avail_info **res,
			  int *nr_modules,
			  int *res_size,
			  int maxdepth)
{
  DIR *directory;
  char *buf = NULL;
  size_t buf_size = 0;
  size_t len;
  struct dirent *entry;
  struct avail_info *info_entry;
  struct stat statbuf;
  size_t dir_skip;
  size_t prefix_skip;
#ifdef HAVE_READLINK
  char *default_buf;
  char *symlink_val = NULL;
  /* symlink_size is the current size of the symlink_val buffer.
     (We only allocate the buffer when we encounter a symlink).  We
     could make this arbitrarily small, since we reallocate the buffer
     if we encounter a large symlink.  Unfortunately, there are
     systems where readlink returns -1 and sets errno to EINVAL
     (PowerMAX OS 4.1 when the symlink is present on a local file
     system is one example).  We don't handle that case, except that
     we use a large enough buffer so that we in practice should never
     have to reallocate it. */
  int symlink_size = 1012;
  int n;
#endif
  
  if (maxdepth <= 0)
    {
      fprintf(stderr, "%s: %s: too deep directory hierarchy\n",
	      argv0, dir);
      return;
    }

  dir_skip = strlen(dir) + 1;

  if (prefix != NULL)
    prefix_skip = dir_skip + strlen(prefix) + 1;
  else
    prefix_skip = dir_skip;
  
  buf_size = prefix_skip + 19;
  buf = xmalloc(buf_size);
  if (prefix != NULL)
    sprintf(buf, "%s/%s", dir, prefix);
  else
    strcpy(buf, dir);

  assert(buf[prefix_skip - 1] == '\0');

  if ((directory = opendir(buf)) == NULL)
    {
      fprintf(stderr, "%s: opening directory ", argv0);
      perror(buf);
      return;
    }

  buf[prefix_skip - 1] = '/';

  while (errno = 0, (entry = readdir(directory)) != NULL)
    {
      len = NAMLEN(entry);
      if (prefix_skip + len + 1 > buf_size)
	{
	  buf_size = prefix_skip + len + 1;
	  buf = xrealloc(buf, buf_size);
	}

      strncpy(buf + prefix_skip, entry->d_name, len);
      buf[prefix_skip + len] = '\0';
      if (is_module_name(buf + prefix_skip))
	{
	  if (stat(buf, &statbuf) != 0)
	    {
	      fprintf(stderr, "%s: while statting ", argv0);
	      perror(buf);
	      exit(CMOD_EXIT_ERROR);
	    }

	  /* Ignore STAT_MACROS_BROKEN here -- if they are, "module
	     avail" simply won't work.  It is enough that the more
	     vital parts of "cmod" tries to work. */
	  if (S_ISDIR(statbuf.st_mode))
	    {
#ifdef HAVE_READLINK
	      default_buf = xmalloc(strlen(buf) + 1 + sizeof(DEFAULT_MODULE));
	      sprintf(default_buf, "%s/%s", buf, DEFAULT_MODULE);
	      if (lstat(default_buf, &statbuf) == 0
		  && S_ISLNK(statbuf.st_mode))
		{
		  if (symlink_val == NULL)
		    symlink_val = xmalloc(symlink_size);
		  while (1)
		    {
		      /* Red Hat Linux 5.0 (and other versions of
			 Linux) will write a NUL character past the
			 buffer, so supply symlink_size-1 instead of
			 symlink_size here.  This will waste one
			 character on all other systems, but better
			 safe than sorry. */

		      n = readlink(default_buf, symlink_val, symlink_size-1);
		      if (n >= 0 && n < symlink_size-1)
			{
			  /* We got the entire symlink, with room to
			     spare.  NUL-terminate it, and exit from
			     this loop.  This is the normal exit from
			     the loop. */
			  symlink_val[n] = '\0';
			  break;
			}
		      else if (n < 0
			       && (errno != ERANGE
#ifdef READLINK_EIO_BROKEN

/* Some versions of Linux (for instance Red Hat 5.0) uses EIO instead
   of ERANGE to signal that the value didn't fit in the supplied
   buffer.  Since EIO is also used for real I/O errors, we have to
   take care not to get stuck in this loop allocating more and more
   memory.  This compromise should be ok. */

				   || (errno != EIO && symlink_size < 10240)
#endif
				   ))
			{
			  fprintf(stderr, "%s: readlink symlink ", argv0);
			  perror(default_buf);
			  exit(1);
			}

		      /* There wasn't room for the symlink.
			 Reallocate and try again. */

		      symlink_size *= 2;
		      symlink_val = xrealloc(symlink_val, symlink_size);
		    }

		  /* OK, we've got the NUL-terminated symlink value
		     in symlink_val. */
		  avail_modules_recursively(dir, buf + dir_skip,
					    symlink_val,
					    res, nr_modules, res_size,
					    maxdepth - 1);

		}
	      else
		{
		  /* The default file, if it exists, is not a symlink. */
		  avail_modules_recursively(dir, buf + dir_skip, NULL,
					    res, nr_modules, res_size,
					    maxdepth - 1);
		}
#else  /* symlinks not supported */
	      avail_modules_recursively(dir, buf + dir_skip, NULL,
					res, nr_modules, res_size,
					maxdepth - 1);
#endif
	    }
	  else
	    {
	      if (*nr_modules >= *res_size)
		{
		  ++*res_size;
		  *res = xrealloc(*res, *res_size * sizeof(struct avail_info));
		}
	      info_entry = &(*res)[(*nr_modules)++];
	      info_entry->name = xstrdup(buf + dir_skip);
	      if (default_module != NULL
		  && !strcmp(buf + prefix_skip, default_module))
		info_entry->is_default = 1;
	      else
		info_entry->is_default = 0;
	    }
	}
    }

  if (errno != 0)
    {
      fprintf(stderr, "%s: reading directory ", argv0);
      perror(dir);
    }

#ifdef CLOSEDIR_VOID
  closedir(directory);
#else

again:
  if (closedir(directory) < 0)
    {
      if (errno == EINTR)
	goto again;
      fprintf(stderr, "%s: closing directory ", argv0);
      perror(dir);
    }
#endif

#ifdef HAVE_READLINK
  xfree(symlink_val);
#endif
}


static int
cmp_avail_infos(const void *a, const void *b)
{
  const struct avail_info *x = (const struct avail_info*)a;
  const struct avail_info *y = (const struct avail_info*)b;

  return strcmp(x->name, y->name);
}


static int
avail_modules(const char *UNUSED(from_module),
	      const int *UNUSED(from_line),
	      int argc,
	      const char *const *UNUSED(argv))
{
  struct env *loaded_modules;
  struct env *module_path;
  int component = 0;
  const char *dir;
  struct avail_info *available_modules = NULL;
  int modules_present;
  int modules_allocated = 0;
  int ix;
  size_t maxlen;
  int cols_used;
  int rows_used;
  int col;
  int row;
  size_t len;

  assert(argc == 0);

  envarray_locate_create(global_env, "LOADEDMODULES", 1);
  module_path = envarray_locate_create(global_env, "MODULEPATH", 1);
  loaded_modules = envarray_locate(global_env, "LOADEDMODULES");

  while ((dir = env_iterate(module_path, ':', &component)) != NULL)
    {
      fprintf(stderr, "\nIn directory %s:\n\n", dir);
      modules_present = 0;
      avail_modules_recursively(dir, NULL, NULL, &available_modules,
				&modules_present, &modules_allocated, 2);
      qsort(available_modules, modules_present, sizeof(struct avail_info),
	    &cmp_avail_infos);
      for (maxlen = 0, ix = 0; ix < modules_present; ++ix)
	{
	  len = (strlen(available_modules[ix].name)
		 + 7 * !!available_modules[ix].is_default);
	  if (len > maxlen)
	    maxlen = len;
	}
      /* FIXME: should not hard-code the screen width.  True, 80
	 characters is the only sane screen width, but the world is
	 full of insane individuals.  For now, they will have to
	 suffer. */
#define SCREEN_WIDTH 80
      /* The format of each module entry is:
         SPACE SPACE PLUS-OR-MINUS NAME DEFopt
	 DEFopt is " (def) " if this is a default module. */
      cols_used = SCREEN_WIDTH / (maxlen + 3);
      if (cols_used < 1)
	cols_used = 1;
      rows_used = (modules_present-1) / cols_used + 1;
      for (row = 0; row < rows_used; ++row)
	{
	  for (col = 0;
	       col < cols_used && row + col*rows_used < modules_present;
	       ++col)
	    {
	      struct avail_info *entry = &available_modules[row+col*rows_used];
	      putc(' ', stderr);
	      putc(' ', stderr);
	      if (env_frag_present(loaded_modules, ':', entry->name))
		putc('+', stderr);
	      else
		putc('-', stderr);
	      if (entry->is_default)
		fprintf(stderr, "%s (def) %*s", entry->name,
			maxlen - strlen(entry->name) - 7, "");
	      else
		fprintf(stderr, "%-*s", maxlen, entry->name);
	    }
	  putc('\n', stderr);
	}
    }

  return 0;
}


static int
cmod_version(const char *UNUSED(from_module),
	     const int *UNUSED(from_line),
	     int UNUSED(argc),
	     const char *const *UNUSED(argv))
{
  fprintf(stderr,
	  "%s version %s.\n\
Copyright (C) 1997-1998 Per Cederqvist, Kent Engstr\366m and \
Henrik Rindl\366w.\n",
	  PACKAGE, VERSION);
  return 0;
}


static const struct cmd cmd_table[] = {
  {"append-path", NULL, 2, 2, append_path, remove_path},
  {"prepend-path", NULL, 2, 2, prepend_path, remove_path},
  {"remove-path", NULL, 2, 2, remove_path, remove_path},
  {"setenv", NULL, 2, 2, set_env, unset_env},
  {"unsetenv", NULL, 1, 1, unset_env, unset_env},
  {"module", "add", 1, -1, add_module, rm_module},
  {"module", "load", 1, -1, add_module, rm_module},
  {"module", "unload", 1, -1, rm_module, add_module},
  {"module", "rm", 1, -1, rm_module, add_module},
  {"module", "use", 1, 1, use_path, unuse_path},
  {"module", "unuse", 1, 1, unuse_path, use_path},
  {"module", "list", 0, 0, list_modules, no_op},
  {"module", "avail", 0, 0, avail_modules, no_op},
  {"module", "--version", 0, 0, cmod_version, no_op},
  {NULL, NULL, 0, 0, NULL, NULL}
};

static void
parse_module(FILE *fp,
	     const char *filename,
	     int for_removal,
	     int *lineno)
{
  /* FIXME: For now, discard the error status. */
  cmod_parse_module(fp, filename, for_removal, cmd_table, lineno);
}


static void
usage(void)
{
  fprintf(stderr,
	  "Usage: module sub-command [args...]\n\n");
  fprintf(stderr,
	  "  Available sub-commands:\n\n");
  fprintf(stderr,
	  "    add|load  modulefile [modulefile...]    Add module\n");
  fprintf(stderr,
	  "    rm|unload modulefile [modulefile...]    Remove module\n");
  fprintf(stderr,
	  "    use       dir                           Add to MODULEPATH\n");
  fprintf(stderr,
	  "    unuse     dir                           Remove from MODULEPATH\
\n");
  fprintf(stderr,
	  "    list                                    List loaded modules\n");
  fprintf(stderr,"\n");
  exit(CMOD_EXIT_ERROR);
}


int
main(int argc,
     const char **argv)
{
  argv0 = argv[0];

#if 0
  /* FIXME: Add a switch to enable selected traces. */
  set_traces(TRACE_ASSIGN);
#endif

  if (argc < 2)
    usage();

  if (!strcmp(argv[1], "sh") || !strcmp(argv[1], "csh"))
    {
    }
  else
    usage();

  global_env = create_envarray();

  switch (cmod_parse_argv(cmd_table, argc-2, argv+2))
    {
    case -1:
      /* bad usage */
      usage();
    case 0:
      /* no error */
      break;
    default:
      /* error (but command line looks ok) */
      break;
    }

  if (cmod_errors > 0)
    exit(CMOD_EXIT_ERROR);

  envarray_dump(global_env, stdout, argv[1]);
  destroy_envarray(global_env);

  if (warning_found)
    exit(CMOD_EXIT_WARNING);
  else
    exit(0);
}
