/* cmod - set up the user environment with module files.
   Copyright (C) 1997 by the authors of cmod (see AUTHORS).

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <sys/types.h>

#include "cmod_parse.h"
#include "cmod_main.h"
#include "cmod_alloc.h"
#include "env.h"
#include "archaic.h"

struct undo_chain;

struct undo_chain {
#ifdef PROTOTYPES
  int (*undo)(const char *module, const int *line,
	      int argc, const char *const *argv);
#else
  int (*undo)();
#endif
  
  int from_line;
  int argc;
  char **argv;
  struct undo_chain *prev;
};


/* File parsing. */
static char *
hunt_for(char terminator,
	 FILE *fp)
{
  char *buf = xmalloc(0);
  int bc = 0;
  int c;

  while ((c = getc(fp)) != EOF && c != '\n' && c != terminator)
    {
      /* FIXME: this is inefficient. */
      buf = xrealloc(buf, ++bc + 1);
      buf[bc-1] = c;
    }

  if (c != terminator)
    {
      /* FIXME: better error handling. */
      fprintf(stderr, "Unexpected end of variable\n");
      exit(CMOD_EXIT_ERROR);
    }

  buf[bc] = '\0';
  return buf;
}


static char *
get_word(FILE *fp,
	 const char *filename,
	 int hunt_cmd,
	 int *lineno)
{
  char *buf = NULL;
  int bc = 0;
  int c;

  errno = 0;

  while (1)
    {
      switch (c = getc(fp))
	{
	case EOF:
	  if (ferror(fp))
	    {
#if defined(STAT_MACROS_BROKEN) && defined(EISDIR)
	      if (errno == EISDIR)
		{
		  fprintf(stderr, "%s: %s lacks %s/default\n", argv0, filename,
			  filename);
		  exit(CMOD_EXIT_ERROR);
		}
#endif
	      fprintf(stderr, "%s: while reading from ", argv0);
	      perror(filename);
	      exit(CMOD_EXIT_ERROR);
	    }

	  return NULL;

	case ' ':
	  continue;

	case '\t':
	  continue;

	case '\n':
	  ++*lineno;
	  if (hunt_cmd)
	    continue;
	  else
	    return NULL;

	case '#':
	  if (hunt_cmd)
	    {
	      while ((c = getc(fp)) != EOF && c != '\n')
		;
	      switch (c)
		{
		case '\n':
		  ++*lineno;
		  continue;
		case EOF:
		  return NULL;
		}
	      assert(0);
	    }
	  else
	    return NULL;
	}
      break;
    }

  do {
    buf = xrealloc(buf, ++bc + 1);
    buf[bc-1] = c;
    c = getc(fp);
  } while (c != EOF && c != ' ' && c != '\t' && c != '\n');

  buf[bc] = '\0';
  ungetc(c, fp);

  return buf;
}


static char *
get_word_expand_variables(FILE *fp,
			  int *lineno)
{
  char *buf = NULL;
  int bc = 0;
  int c;

  while (1)
    {
      switch (c = getc(fp))
	{
	case EOF:
	  /* FIXME: should check for errors. */
	  return NULL;

	case ' ':
	  continue;

	case '\t':
	  continue;

	case '\n':
	  ++*lineno;
	  return NULL;

	case '#':
	  return NULL;
	}
      break;
    }

  do {
    if (c == '$')
      {
	char *var;
	struct env *val_env;
	const char *val = NULL;
	int len;

	c = getc(fp);
	if (c == 'e')
	  {
	    if (getc(fp) == 'n' && getc(fp) == 'v' && getc(fp) == '(')
	      var = hunt_for(')', fp);
	    else
	      {
		/* FIXME: improve error reporting. */
		fprintf(stderr, "%s: bad $ substitution\n", argv0);
		exit(CMOD_EXIT_ERROR);
	      }
	  }
	else if (c == '{')
	  var = hunt_for('}', fp);
	else
	  {
	    /* FIXME: improve error reporting. */
	    fprintf(stderr, "%s: missing $ substitution\n", argv0);
	    exit(CMOD_EXIT_ERROR);
	  }
	val_env = envarray_locate_create(global_env, var, -1);
	if (val_env == NULL)
	  {
	    /* FIXME: improve error reporting -- print module name. */
	    fprintf(stderr, "%s: missing variable %s\n", argv0, var);
	    exit(CMOD_EXIT_ERROR);
	  }
	val = env_value(val_env);
	len = strlen(val);
	buf = xrealloc(buf, bc + len + 1);
	strcpy(buf + bc, val);
	bc += len;
      }
    else
      {
	buf = xrealloc(buf, ++bc + 1);
	buf[bc-1] = c;
      }
    c = getc(fp);
  } while (c != EOF && c != ' ' && c != '\t' && c != '\n');

  buf[bc] = '\0';
  ungetc(c, fp);

  return buf;
}

static const struct cmd *
locate_match(FILE *fp,
	     const char *filename,
	     const char *w,
	     const char *arg_1,
	     const struct cmd *parse_table,
	     int *lineno)
{
  const struct cmd *attempt;

  assert(fp == NULL ? lineno == NULL : lineno != NULL);
  assert(w != NULL);
  assert(fp == NULL ? 1 : arg_1 == NULL);

  for (attempt = parse_table; attempt->cmd1 != NULL; ++attempt)
    {
      /* Does the first word match? */
      if (strcmp(w, attempt->cmd1))
	continue;

      /* Must the second word match as well? */
      if (attempt->cmd2 != NULL)
	{
	  /* Yes, the second word must match.  Store one word of
	     lookahead in arg_1. */

	  if (arg_1 == NULL && fp != NULL)
	    arg_1 = get_word(fp, filename, 0, lineno);
	  if (arg_1 == NULL)
	    return NULL;
	  if (strcmp(arg_1, attempt->cmd2))
	    continue;
	}

      /* We have found a match.  */
      return attempt;
    }

  return NULL;
}


/* FIXME: should have better error messages than this. */
static void
syntax_error(const char *filename,
	     const int *lineno)
{
  if (cmod_errors > 20)
    return;

  ++cmod_errors;

  fprintf(stderr, "%s:%d: syntax error\n", filename, *lineno);
}

int
cmod_parse_module(FILE *fp,
		  const char *filename,
		  int for_removal,
		  const struct cmd *parse_table,
		  int *lineno)
{
  char *w;
  int i;
  struct undo_chain *undo = NULL;
  const struct cmd *match;
  int argc;
  int arg_alloc;
  char **argv;
  int err_found = 0;

  while ((w = get_word(fp, filename, 1, lineno)) != NULL)
    {
      match = locate_match(fp, filename, w, NULL, parse_table, lineno);

      if (match == NULL)
	{
	  syntax_error(filename, lineno);
	  err_found = 1;
	}
      
      /* Parse arguments until end of line. */
      argc = 0;
      arg_alloc = match->min_args;
      assert(arg_alloc > 0);
      argv = xmalloc(arg_alloc * sizeof(char*));

      while ((w = get_word_expand_variables(fp, lineno)) != NULL)
	{
	  if (argc >= arg_alloc)
	    {
	      char **next = xrealloc(argv, ++arg_alloc * sizeof(char*));
	      argv = next;
	    }
	  argv[argc++] = w;
	}

      /* Check argument count. */
      if (argc < match->min_args)
	{
	  syntax_error(filename, lineno);
	  err_found = 1;
	  goto free_argv;
	}

      if (match->max_args != -1 && argc > match->max_args)
	{
	  syntax_error(filename, lineno);
	  err_found = 1;
	  goto free_argv;
	}
	    
      /* We have the arguments.  Perform the command. */
      /* FIXME: ignore the error status for now. */
      /* FIXME: why is the cast below necessary? */
      (*match->doit)(filename, lineno, argc, (const char*const*)argv);

      if (for_removal)
	{
	  struct undo_chain *head = xmalloc(sizeof(struct undo_chain));

	  head->undo = match->undo;
	  head->argc = argc;
	  head->argv = argv;
	  head->from_line = *lineno;
	  head->prev = undo;
	  undo = head;
	  goto free_arg1;
	}
      else
	{
	free_argv:
	  for (i = 0; i < argc; ++i)
	    xfree(argv[i]);
	  xfree(argv);
	}

    free_arg1:
      xfree(w);
    }

  while (undo != NULL)
    {
      struct undo_chain *prev = undo->prev;
      assert(for_removal);
      /* FIXME: why is the cast below necessary? */
      (*undo->undo)(filename, &undo->from_line,
		    undo->argc, (const char*const*)undo->argv);
      for (i = 0; i < undo->argc; ++i)
	xfree(undo->argv[i]);
      xfree(undo->argv);
      xfree(undo);
      undo = prev;
    }

  return err_found;
}

int
cmod_parse_argv(const struct cmd *parse_table,
		int argc,
		const char **argv)
{
  const struct cmd *match;

  if (argc < 1)
    {
      fprintf(stderr, "missing subcommand\n");
      return -1;
    }

  match = locate_match(NULL, NULL, "module", argv[0], parse_table, NULL);
  if (match == NULL)
      return -1;

  if (argc - 1 < match->min_args) 
    {
      fprintf(stderr, "missing argument to %s%s%s\n", match->cmd1,
	      match->cmd2 ? " ": "", match->cmd2 ? match->cmd2: "");
      return -1;
    }

  if (match->max_args != -1 && argc - 1 > match->max_args)
    {
      fprintf(stderr, "too many arguments to %s%s%s\n", match->cmd1,
	      match->cmd2 ? " ": "", match->cmd2 ? match->cmd2: "");
      return -1;
    }

  return (*match->doit)(NULL, NULL, argc - 1, argv + 1);
}

/* cmod_parse.c ends here */
