/* cmod - set up the user environment with module files.
   Copyright (C) 1997 by the authors of cmod (see AUTHORS).

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <assert.h>
#include <stdio.h>
#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif
#include <string.h>
#include <sys/types.h>

#include "env.h"
#include "cmod_main.h"
#include "cmod_trace.h"
#include "cmod_alloc.h"
#include "archaic.h"

struct env {
  /* Name of the environment variable. */
  char *name;

  /* True if the environment variable is explicitly unset. */
  int unset;

  /* True if the environment variable has been read from the real
     environment and not modified since. */
  int clean;

  /* The current delimiter used to separate the fragments in frags.
     This is only valid when nfrags is non-zero. */
  char current_delimiter;

  /* Number of fragments, or 0 if frags is not valid. */
  int nfrags;

  /* The fragments, each allocated with xmalloc.  Non-fragmented
     variables are stored as a single fragment. */
  char **frags;
  
  /* The unfragmented value, or NULL if non_fragmented isn't valid. */
  char *non_fragmented;
};


struct envarray {
  /* Number of objects in ARR. */
  int size;

  /* The list of variables. */
  struct env *arr;
};


/* Check that the "struct env" is sane. */
static void
env_sane(struct env *li, int relax)
{
  int i;
  char *n;
  char *f;

  if (!relax)
    assert(li->name != NULL);	/* May be CMOD_NULL. */
  assert(li->clean == 0 || li->clean == 1);
  assert(li->unset == 0 || li->unset == 1);
  if (li->unset)
    {
      assert(li->nfrags == 0);
      assert(li->frags == NULL);
      assert(li->non_fragmented == NULL);
    }
  else
    {
      if (li->nfrags > 0)
	{
	  assert(li->current_delimiter != '\0');
	  assert(li->frags != NULL && li->frags != CMOD_NULL);
	  if (!relax)
	    for (i = 0; i < li->nfrags; ++i)
	      assert(li->frags[i] != NULL && li->frags[i] != CMOD_NULL);
	}
      else
	assert(li->frags == NULL || li->frags == CMOD_NULL);

      if (li->non_fragmented != NULL && li->nfrags > 0)
	{
	  /* Make sure that both representations contain the same string. */
	  n = li->non_fragmented;
	  for (i = 0; i < li->nfrags; ++i)
	    {
	      if (relax && li->frags[i] == NULL)
		/* We sometimes steal memory. */
		break;

	      f = li->frags[i];
	      while (*f != '\0' && *n != '\0')
		{
		  assert(*f != li->current_delimiter);
		  assert(*f == *n);
		  ++f;
		  ++n;
		}
	      assert(*f == '\0');
	      if (i < li->nfrags - 1)
		{
		  assert(*n == li->current_delimiter);
		  ++n;
		}
	      else
		assert(*n == '\0');
	    }
	}
    }
}

  
void
env_ctor(struct env *li)
{
  li->name = NULL;
  li->unset = 0;
  li->clean = 0;
  li->current_delimiter = 0;
  li->nfrags = 0;
  li->frags = NULL;
  li->non_fragmented = NULL;
  env_sane(li, 1);
}


void
env_dtor(struct env *li)
{
  int i;

  env_sane(li, 1);

  if (li->name)
    xfree(li->name);
  for (i = 0; i < li->nfrags; ++i)
    if (li->frags[i])
      xfree(li->frags[i]);
  if (li->frags != NULL)
    xfree(li->frags);
  if (li->non_fragmented != NULL)
    xfree(li->non_fragmented);
}


static void
env_invalidate_unfragmented(struct env *li)
{
  if (li->non_fragmented != NULL)
    {
      xfree(li->non_fragmented);
      li->non_fragmented = NULL;
    }
}

static void
env_invalidate_fragmented(struct env *li)
{
  int i;

  for (i = 0; i < li->nfrags; ++i)
    if (li->frags[i])
      xfree(li->frags[i]);
  if (li->frags != NULL)
    xfree(li->frags);
  li->current_delimiter = '\0';
  li->nfrags = 0;
  li->frags = NULL;
}


static void
env_fragment(struct env *li,
	     char frag_delimiter)
{
  int nfrags = 0;
  char *frag;
  char *frag_end;
  char *copy;

  if (li->nfrags > 0 && li->current_delimiter == frag_delimiter)
    return;

  env_value(li);

  env_invalidate_fragmented(li);

  li->current_delimiter = frag_delimiter;

  if (li->non_fragmented == NULL || *li->non_fragmented == '\0')
    return;
  
  for (frag = li->non_fragmented;
       frag != NULL;
       frag = strchr(frag, frag_delimiter))
    {
      ++nfrags;
      ++frag;
    }

  assert(nfrags > 0);
  li->frags = xmalloc(nfrags * sizeof(char*));
  li->nfrags = nfrags;

  frag = li->non_fragmented;
  assert(frag != NULL);

  for (nfrags = 0; nfrags < li->nfrags; ++nfrags)
    {
      frag_end = strchr(frag, frag_delimiter);
      if (frag_end == NULL)
	{
	  assert(nfrags == li->nfrags - 1);
	  copy = xstrdup(frag);
	}
      else
	{
	  assert(nfrags < li->nfrags - 1);
	  copy = xmalloc((frag_end - frag) + 1);
	  strncpy(copy, frag, frag_end - frag);
	  copy[frag_end - frag] = '\0';
	}

      li->frags[nfrags] = copy;
      frag = frag_end + 1;
    }

  assert(nfrags == li->nfrags);
}


int
env_frag_present(struct env *li,
		 char frag_delimiter,
		 const char *frag)
{
  int i;

  env_sane(li, 0);
  env_fragment(li, frag_delimiter);
  env_sane(li, 0);

  for (i = 0; i < li->nfrags; ++i)
    if (!strcmp(li->frags[i], frag))
      return 1;
  return 0;
}


void
env_append_one_frag(struct env *li,
		    char frag_delimiter,
		    const char *frag)
{
  assert(frag != NULL);
  assert(li != NULL);

  env_sane(li, 0);
  env_fragment(li, frag_delimiter);

  li->frags = xrealloc(li->frags, ++li->nfrags * sizeof(char*));
  li->frags[li->nfrags-1] = xstrdup(frag);
  li->unset = 0;
  li->clean = 0;

  env_invalidate_unfragmented(li);
  env_sane(li, 0);

  trace(TRACE_ASSIGN, "appending %s << %s\n", li->name, frag);
}


void
env_remove_one_frag(struct env *li,
		    char frag_delimiter,
		    const char *frag)
{
  int i;
  int move_i;

  assert(frag != NULL);
      
  env_sane(li, 0);
  env_fragment(li, frag_delimiter);

  for (i = 0; i < li->nfrags; )
    if (!strcmp(li->frags[i], frag))
      {
	trace(TRACE_ASSIGN, "removing %s >> %s\n", li->name, frag);
	env_invalidate_unfragmented(li);
	for (move_i = i; move_i < li->nfrags - 1; move_i++)
	  li->frags[move_i] = li->frags[move_i + 1];

	li->frags = xrealloc(li->frags, --li->nfrags * sizeof(char*));
	li->clean = 0;
      }
    else
      ++i;

  env_sane(li, 0);
}


void
env_append_frags(struct env *li,
		 char frag_delimiter,
		 const char *frags,
		 int unique)
{
  char *frag_end;
  char *copy;

  assert(frag_delimiter != '\0');

  env_sane(li, 0);

  env_fragment(li, frag_delimiter);

  /* Ignore attempts to append an empty string. */
  if (*frags == '\0')
    {
      env_sane(li, 0);
      return;
    }

  while (1)
    {
      frag_end = strchr(frags, frag_delimiter);
      if (frag_end == NULL)
	{
	  if (!unique || !env_frag_present(li, frag_delimiter, frags))
	    {
	      copy = xstrdup(frags);
	      env_append_one_frag(li, frag_delimiter, copy);
	    }
	  env_sane(li, 0);
	  return;
	}
      else
	{
	  copy = xmalloc(frag_end - frags + 1);
	  strncpy(copy, frags, frag_end - frags);
	  copy[frag_end - frags] = '\0';
	  if (!unique || !env_frag_present(li, frag_delimiter, copy))
	    env_append_one_frag(li, frag_delimiter, copy);
	  frags = frag_end + 1;
	}
    }
}


void
env_prepend_one_frag(struct env *li,
		     char frag_delimiter,
		     char *frag)
{
  int i;

  trace(TRACE_ASSIGN, "prepending %s << %s\n", li->name, frag);

  env_sane(li, 0);

  env_fragment(li, frag_delimiter);
  env_invalidate_unfragmented(li);

  li->frags = xrealloc(li->frags, ++li->nfrags * sizeof(char*));

  for (i = li->nfrags - 2; i >= 0; i--)
    li->frags[i + 1] = li->frags[i];
	
  li->frags[0] = frag;
  li->unset = 0;
  li->clean = 0;
  env_sane(li, 0);
}


void
env_prepend_frags(struct env *li,
		  char frag_delimiter,
		  const char *frags,
		  int unique)
{
  int fragno;
  struct env parsed_frag;

  env_sane(li, 0);

  env_ctor(&parsed_frag);
  parsed_frag.name = xmalloc(0);
  env_append_frags(&parsed_frag, frag_delimiter, frags, unique);
  env_sane(&parsed_frag, 0);

  for (fragno = parsed_frag.nfrags - 1; fragno >= 0 ; fragno--)
    {
      if (unique)
	env_remove_one_frag(li, frag_delimiter, parsed_frag.frags[fragno]);

      /* Steal the memory from parsed_frag. */
      env_prepend_one_frag(li, frag_delimiter, parsed_frag.frags[fragno]);
      /* Make parsed_frag forget about the stolen memory. */
      parsed_frag.frags[fragno] = NULL;
    }

  env_dtor(&parsed_frag);
  env_sane(li, 0);
}


void
env_remove_frags(struct env *li,
		 char frag_delimiter,
		 const char *frags)
{
  char *frag_end;
  char *copy;

  env_sane(li, 0);
  assert(frag_delimiter != '\0');

  while (1)
    {
      frag_end = strchr(frags, frag_delimiter);
      if (frag_end == NULL)
	{
	  env_remove_one_frag(li, frag_delimiter, frags);
	  env_sane(li, 0);
	  return;
	}
      else
	{
	  copy = xmalloc(frag_end - frags + 1);
	  strncpy(copy, frags, frag_end - frags);
	  copy[frag_end - frags] = '\0';
	  env_remove_one_frag(li, frag_delimiter, copy);
	  frags = frag_end + 1;
	  xfree(copy);
	}
    }
}


void
env_setenv(struct env *li,
	   char *val)
{
  trace(TRACE_ASSIGN, "setting %s == %s\n", li->name, val);

  env_sane(li, 0);
  env_invalidate_fragmented(li);
  if (li->non_fragmented != NULL)
    xfree(li->non_fragmented);
  li->non_fragmented = val;
  li->unset = 0;
  li->clean = 0;
  env_sane(li, 0);
}


void
env_unsetenv(struct env *li)
{
  trace(TRACE_ASSIGN, "unsetting %s\n", li->name);

  env_sane(li, 0);
  env_invalidate_fragmented(li);
  env_invalidate_unfragmented(li);
  li->unset = 1;
  li->clean = 0;
  env_sane(li, 0);
}

void
env_dump(struct env *li,
	 FILE *f,
	 const char *sh)
{
  env_sane(li, 0);

  env_value(li);

  if (!strcmp(sh, "sh"))
    {
      if (li->unset)
	fprintf(f, "unset %s;\n", li->name);
      else
	{
	  fprintf(f, "%s=", li->name);
	  /* FIXME: quote special characters */
	  fprintf(f, "%s", li->non_fragmented);
	  fprintf(f, ";\nexport %s;\n", li->name);
	}
    }
  else if (!strcmp(sh, "csh"))
    {
      if (li->unset)
	fprintf(f, "unsetenv %s;\n", li->name);
      else
	{
	  fprintf(f, "setenv %s ", li->name);
	  /* FIXME: quote special characters */
	  fprintf(f, "%s", li->non_fragmented);
	  fputs(";\n", f);
	}
    }
  else
    {
      fprintf(stderr, "%s: Unsupported shell \"%s\"", argv0, sh);
      exit(1);
    }
  env_sane(li, 0);
}


const char *
env_iterate(struct env *li,
	    char frag_delimiter,
	    int *iterator)
{
  env_sane(li, 0);

  if (*iterator == 0)
    {
      env_fragment(li, frag_delimiter);
      env_sane(li, 0);
    }
  else
    {
      assert(li->current_delimiter == frag_delimiter);
    }
  
  if (*iterator < li->nfrags)
    return li->frags[(*iterator)++];
  else
    return NULL;
}

/* This defragments the variable as a side-effect. */
const char *
env_value(struct env *li)
{
  size_t size;
  int i;
  char *tail;
  char *end;

  env_sane(li, 0);

  if (li->non_fragmented != NULL)
    return li->non_fragmented;

  if (li->unset)
    return NULL;

  if (li->nfrags == 0)
    {
      li->non_fragmented = xmalloc(1);
      li->non_fragmented[0] = '\0';
      env_sane(li, 0);
      return li->non_fragmented;
    }

  assert(li->nfrags > 0);

  size = li->nfrags;   /* Allow room for separators and terminating NUL. */

  for (i = 0; i < li->nfrags; ++i)
    size += strlen(li->frags[i]);

  li->non_fragmented = xmalloc(size);
  tail = li->non_fragmented;
  for (i = 0; i < li->nfrags; ++i)
    {
      if (i > 0)
	*tail++ = li->current_delimiter;
      end = strchr(li->frags[i], '\0');
      assert(tail + (end - li->frags[i]) < li->non_fragmented + size);
      strncpy(tail, li->frags[i], end - li->frags[i]);
      tail += end - li->frags[i];
    }
  *tail++ = '\0';
  assert(tail == li->non_fragmented + size);

  env_sane(li, 0);

  return li->non_fragmented;
}


struct env *
envarray_locate(struct envarray* arr,
		const char* name)
{
  int i;

  for (i = 0; i < arr->size; ++i)
    if (!strcmp(name, arr->arr[i].name))
      return &arr->arr[i];

  return NULL;
}

struct env *
envarray_locate_create(struct envarray* arr,
			const char* name,
			int read_env)
{
  char *current_env = NULL;
  struct env *entry;

  if ((entry = envarray_locate(arr, name)) != NULL)
    return entry;

  if (read_env)
    current_env = getenv(name);

  /* Fail if no value is available and read_env is -1. */
  if (read_env == -1 && current_env == NULL)
    return NULL;

  arr->arr = xrealloc(arr->arr, ++arr->size * sizeof(struct env));

  entry = &arr->arr[arr->size-1];
  
  env_ctor(entry);

  entry->name = xstrdup(name);

  if (current_env != NULL)
    {
      current_env = xstrdup(current_env);
      env_setenv(entry, current_env);
      entry->clean = 1;
    }

  env_sane(entry, 0);

  return entry;
}


void
envarray_dump(struct envarray *arr,
	      FILE *f,
	      const char *sh)
{
  int i;

  for (i = 0; i < arr->size; i++)
    if (!arr->arr[i].clean)
      env_dump(&arr->arr[i], f, sh);
}


struct envarray *
create_envarray(void)
{
  struct envarray *result;

  result = xmalloc(sizeof(struct envarray));
  result->size = 0;
  result->arr = NULL;
  return result;
}


void
destroy_envarray(struct envarray *arr)
{
  int i;

  for (i = 0; i < arr->size; i++)
    env_dtor(&arr->arr[i]);
  xfree(arr->arr);
  xfree(arr);
}
